"""Structured logger that sends logs to AWS CloudWatch."""

import contextlib
import logging
import time
from threading import Lock, Timer
from typing import Any, Dict, List, Optional

import boto3
import botocore.errorfactory
import botocore.exceptions

from opensearch_log import json_log
from opensearch_log.base_handler import BaseHandler
from opensearch_log.stdout_handler import add_stdout_json_handler

BUFFER_SIZE = 100
FLUSH_SECONDS = 3.0
RETRY_NUM = 3


class CloudwatchHandler(BaseHandler):  # pylint: disable=too-many-instance-attributes
    """Handler that sends log records to AWS CloudWatch."""

    def __init__(self, *args: Any, log_group: str, log_stream: str, **kwargs: Any) -> None:
        """Initialize the handler."""
        super().__init__(*args, **kwargs)
        self._log_client = None
        self.log_group = log_group
        self.log_stream = log_stream
        self.buffer_size = BUFFER_SIZE
        self.flush_seconds = FLUSH_SECONDS
        self._buffer: List[Dict[str, Any]] = []
        self._buffer_lock: Lock = Lock()
        self._timer: Optional[Timer] = None

        self._initialize_log_group()
        self._initialize_log_stream()

    def _initialize_log_stream(self) -> None:
        """Create the log stream if it doesn't already exist."""
        with contextlib.suppress(self.log_client.exceptions.ResourceAlreadyExistsException):
            self.log_client.create_log_stream(
                logGroupName=self.log_group, logStreamName=self.log_stream
            )

    def _initialize_log_group(self) -> None:
        """Create the log group if it doesn't already exist.

        Ignore access deny just in case the group really exists,
        but we do not have permission to create it and so the command will fail.
        If the group does not exist, and we fail to create it, in posting log messages we get
        more understandable error "The specified log group does not exist".
        """
        try:
            self.log_client.create_log_group(logGroupName=self.log_group)
        except self.log_client.exceptions.ResourceAlreadyExistsException:
            # Log group already exists. No action needed.
            pass
        except botocore.exceptions.ClientError as exc:
            # Check if the error is access denied error
            if exc.response["Error"]["Code"] != "AccessDeniedException":
                # Reraise the exception if it's not an AccessDeniedException
                raise

    @property
    def log_client(self) -> boto3.client:
        """Get the boto3 client for CloudWatch logs."""
        if self._log_client is None:
            # Disable boto's built-in logging to avoid recursion
            boto3.set_stream_logger("boto3", logging.CRITICAL)
            boto3.set_stream_logger("botocore", logging.CRITICAL)

            self._log_client = boto3.client("logs")
        return self._log_client

    def send_message(self, message: Optional[str], record: logging.LogRecord) -> None:
        """Buffer the log message and flush if necessary."""
        timestamp = int(round(time.time() * 1000))
        log_event = {"timestamp": timestamp, "message": message}

        with self._buffer_lock:
            self._buffer.append(log_event)

        if len(self._buffer) >= self.buffer_size:
            self.flush()
        else:
            self._schedule_flush()

    def flush(self) -> None:
        """Flush the buffer to CloudWatch.

        Could run from Timer's thread or from the main thread.
        So we should not use json_formatter's routines that access ThreadLocal variables.
        """
        if hasattr(self, "_timer") and self._timer is not None and self._timer.is_alive():
            self._timer.cancel()
        self._timer = None

        if not self._buffer:
            return

        with self._buffer_lock:
            logs_buffer = self._buffer
            self._buffer = []

        retries = 0
        while retries < RETRY_NUM:
            try:
                self.log_client.put_log_events(
                    logGroupName=self.log_group,
                    logStreamName=self.log_stream,
                    logEvents=logs_buffer,
                )
                break
            except botocore.exceptions.ClientError as e:
                retries += 1
                print(f"Retry {retries}: Could not send logs to CloudWatch: {e}")
                if retries >= RETRY_NUM:
                    print("Exhausted retries. Lost logs:\n", logs_buffer)

    def _schedule_flush(self) -> None:
        """Schedule a flush operation."""
        if self._timer is None:
            self._timer = Timer(self.flush_seconds, self.flush)
            self._timer.daemon = True
            self._timer.start()

    def close(self) -> None:
        """Flush the buffer and release any outstanding resource."""
        self.flush()
        super().close()


def get_logger(
    *args: Any,
    echo_stdout: bool = False,
    log_group: str,
    log_stream: str,
    log_handler: Optional[BaseHandler] = None,
    **kwargs: Any,
) -> logging.Logger:
    """Create a logger that stream logs to CloudWatch."""
    assert log_handler is None, "log_handler should not be specified"
    logger = json_log.get_logger(
        *args,
        log_handler=CloudwatchHandler(log_group=log_group, log_stream=log_stream),
        **kwargs,
    )
    if echo_stdout:
        add_stdout_json_handler(logger)
    return logger



"""Structured logger that sends logs to AWS OpenSearch."""

import logging
import traceback
from datetime import datetime, timezone
from enum import Enum
from threading import Lock, Timer
from typing import Any, Dict, List, Optional, Union

import boto3
from opensearchpy import AWSV4SignerAuth, OpenSearch, RequestsHttpConnection, helpers

from opensearch_log import json_log
from opensearch_log.base_handler import BaseHandler
from opensearch_log.opensearch_serializer import OpenSearchSerializer
from opensearch_log.stdout_handler import add_stdout_json_handler


class IndexRotation(Enum):
    """Index rotation frequency."""

    DAILY = 0
    MONTHLY = 10


DEFAULT_OPENSEARCH_HOST = "localhost"
DEFAULT_INDEX_NAME = "myindex"
BUFFER_SIZE = 200  # number of records to buffer before sending to OpenSearch
BULK_SIZE = 15  # number of records to send to OpenSearch in one bulk
FLUSH_SECONDS = 0.8  # number of seconds to wait before flushing the buffer
RETRY_NUM = 3  # number of retries to send logs to OpenSearch
INDEX_DATE_FORMAT = "%Y.%m.%d"
IGNORED_LOG_RECORD_FIELDS = [
    "args",
    "levelno",
    "pathname",
    "relativeCreated",
    "msecs",
    "exc_text",
]


class OpensearchHandler(BaseHandler):  # pylint: disable=too-many-instance-attributes
    """Handler to send log records to AWS OpenSearch."""

    DAILY = IndexRotation.DAILY
    MONTHLY = IndexRotation.MONTHLY

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *args: Any,
        opensearch_host: str = DEFAULT_OPENSEARCH_HOST,
        index_name: str = DEFAULT_INDEX_NAME,
        index_rotate: Union[IndexRotation, str] = IndexRotation.DAILY,
        buffer_size: int = BUFFER_SIZE,
        flush_seconds: float = FLUSH_SECONDS,
        **kwargs: Any,
    ) -> None:
        """Initialize the handler."""
        super().__init__(*args, **kwargs)

        # Opensearch
        self.opensearch_host = opensearch_host
        self.index_name = index_name
        if isinstance(index_rotate, str):
            self.index_rotate = IndexRotation[index_rotate]
        else:
            self.index_rotate = index_rotate

        # Bulk settings
        self.buffer_size = buffer_size
        self.flush_seconds = flush_seconds

        # Internals
        self._client: Optional[OpenSearch] = None
        self._buffer: List[Dict[str, Any]] = []
        self._buffer_lock: Lock = Lock()
        self._timer: Optional[Timer] = None

        if self.is_connected():
            print("Connected to OpenSearch")
        else:
            print("Could not connect to OpenSearch instance")

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a logging record in a processed format."""
        self.send_message(None, record)

    def send_message(self, message: Optional[str], record: logging.LogRecord) -> None:
        """Send the log message to OpenSearch."""
        doc = self._convert_log_record_to_opensearch_doc(record)
        with self._buffer_lock:
            self._buffer.append(doc)

        if len(self._buffer) >= self.buffer_size:
            self.flush()
        else:
            self._schedule_flush()

    def is_connected(self) -> bool:
        """Return True if we can ping the OpenSearch."""
        if self._get_opensearch_client() is None:
            return True  # test mode?
        return self._get_opensearch_client().ping()  # type: ignore

    def flush(self) -> None:
        """Flush the buffer."""
        if hasattr(self, "_timer") and self._timer is not None and self._timer.is_alive():
            self._timer.cancel()
        self._timer = None

        if not self._buffer:
            return

        with self._buffer_lock:
            logs_buffer = self._buffer
            self._buffer = []

        index = self._get_index_name()
        actions = [{"_index": index, "_source": record} for record in logs_buffer]

        if self._get_opensearch_client() is None:
            print(
                "Mock MODE: would NOT send to OpenSearch"
            )  # do not use logger here, to avoid recursion
        else:
            for i in range(0, len(actions), BULK_SIZE):
                retries = 0
                while retries < RETRY_NUM:
                    try:
                        helpers.bulk(
                            client=self._get_opensearch_client(),
                            actions=actions[i : i + BULK_SIZE],
                            stats_only=True,
                        )
                        break
                    except Exception as exception:  # pylint: disable=broad-except
                        retries += 1  # Increment retry count
                        print(f"Retry {retries}: Could not send logs to OpenSearch: {exception}")
                        if retries >= RETRY_NUM:
                            print("Exhausted retries. Lost logs:\n", logs_buffer)
                            print(
                                f"\n-- opensearch.bulk() --\n{traceback.format_exc()}-- end --\n"
                            )

    def close(self) -> None:
        """Flush the buffer on close."""
        self.flush()

    def _get_opensearch_client(self) -> Optional[OpenSearch]:
        if self._client is None:
            print(f"Connecting to OpenSearch instance: {self.opensearch_host}...")
            credentials = boto3.Session().get_credentials()
            region = boto3.Session().region_name
            if region is None:
                raise ValueError(
                    "AWS region is not set.\n"
                    "Please set the AWS_REGION environment variable or `region=` in AWS config."
                )
            awsauth = AWSV4SignerAuth(credentials, region, "es")
            verify = self.opensearch_host.startswith("https")
            self._client = OpenSearch(
                [self.opensearch_host],
                http_auth=awsauth,
                use_ssl=verify,
                verify_certs=verify,
                connection_class=RequestsHttpConnection,
                serializer=OpenSearchSerializer(),
            )
            if not self._client.ping():
                print(f"Could not connect to OpenSearch instance at: {self.opensearch_host}...")
            else:
                print(f"Connected to OpenSearch instance at: {self.opensearch_host}...")
        return self._client

    def _schedule_flush(self) -> None:
        if self._timer is None:
            self._timer = Timer(self.flush_seconds, self.flush)
            self._timer.daemon = True
            self._timer.start()

    def _get_index_name(self) -> str:
        if IndexRotation.DAILY == self.index_rotate:
            return self._get_index_name_daily()
        if IndexRotation.MONTHLY == self.index_rotate:
            return self._get_index_name_monthly()
        raise ValueError(f"Unknown index rotation: {self.index_rotate}")

    def _convert_log_record_to_opensearch_doc(self, record: logging.LogRecord) -> Dict[str, Any]:
        """Convert logging.LogRecord to OpenSearch doc."""
        print(f"record: {record.msg}, {record.__dict__}")
        log_record_dict = record.__dict__.copy()
        doc = {
            "@timestamp": self._get_opensearch_datetime_str(
                datetime.now(timezone.utc).timestamp()
            ),
        }
        for key, value in log_record_dict.items():
            if key not in IGNORED_LOG_RECORD_FIELDS:
                doc[key] = "" if value is None else value
        return doc

    def _get_index_name_daily(self, current_date: Optional[datetime] = None) -> str:
        if current_date is None:
            current_date = datetime.now(tz=timezone.utc)  # pragma: no cover
        return f"{self.index_name}-{current_date.strftime(INDEX_DATE_FORMAT)}"

    def _get_index_name_monthly(self, current_date: Optional[datetime] = None) -> str:
        if current_date is None:
            current_date = datetime.now(tz=timezone.utc)  # pragma: no cover
        first_date_of_month = datetime(current_date.year, current_date.month, 1)
        return f"{self.index_name}-{first_date_of_month.strftime(INDEX_DATE_FORMAT)}"

    @staticmethod
    def _get_opensearch_datetime_str(timestamp: float) -> str:
        datetime_utc = datetime.utcfromtimestamp(timestamp)
        fmt = "%Y-%m-%dT%H:%M:%S"
        return f"{datetime_utc.strftime(fmt)}.{int(datetime_utc.microsecond / 1000):03d}Z"


def get_logger(  # pylint: disable=too-many-arguments
    *args: Any,
    opensearch_host: str = DEFAULT_OPENSEARCH_HOST,
    index_name: str = DEFAULT_INDEX_NAME,
    index_rotate: Union[IndexRotation, str] = IndexRotation.DAILY,
    echo_stdout: bool = False,
    buffer_size: int = BUFFER_SIZE,
    flush_seconds: float = FLUSH_SECONDS,
    log_handler: Optional[BaseHandler] = None,
    **kwargs: Any,
) -> logging.Logger:
    """Create a logger that stream logs to OpenSearch."""
    assert log_handler is None, "log_handler should not be specified"
    result = json_log.get_logger(
        *args,
        log_handler=OpensearchHandler(
            opensearch_host=opensearch_host,
            index_name=index_name,
            index_rotate=index_rotate,
            buffer_size=buffer_size,
            flush_seconds=flush_seconds,
        ),
        **kwargs,
    )
    if echo_stdout:
        add_stdout_json_handler(result)
    logging.getLogger("opensearch").setLevel(logging.WARNING)  # suppress HTTP tracing
    return result


if __name__ == "__main__":
    from opensearch_log import Logging

    # from stdout_handler import get_logger

    logger = get_logger(echo_stdout=True)
    with Logging(my_log_field="From Python"):
        logger.info("Hello World")



"""JSON serializer."""
from typing import Any

from opensearchpy.serializer import JSONSerializer


class OpenSearchSerializer(JSONSerializer):
    """Override OpenSearch JSON serializer.

    Ignore serialization errors.
    """

    def default(self, data: Any) -> Any:
        """Catch all serialization fails and fall to __str__."""
        try:
            return super().default(data)
        except TypeError:
            return str(data)



import io
import json
import logging
import os
from contextlib import contextmanager
from typing import Optional, Any, Dict, List, Tuple
from unittest.mock import patch, MagicMock

from opensearch_log.json_log import _log_values
from opensearch_log import json_log
from opensearch_log.cloudwatch_handler import  CloudwatchHandler
from moto import mock_logs

import boto3

from opensearch_log.stdout_handler import get_logger
from opensearch_log.base_handler import BaseHandler
from opensearch_log.stdout_handler import StdoutHandler

import pytest

from opensearch_log.opensearch_handler import OpensearchHandler
from opensearch_log.opensearch_serializer import OpenSearchSerializer
import opensearchpy.exceptions


@pytest.fixture(scope="function")
def log():
    return get_logger(application="-mock-application-")


@contextmanager
def capture_logs(log):
    """Fixture to capture output of send_message in log handlers."""
    captured_messages = io.StringIO()

    def new_send_message(self, message: Optional[str], record: logging.LogRecord):
        """Send message to captured_messages."""
        captured_messages.write(message + "\n")  # Added newline for easier parsing later

    handlers_to_patch = [handler for handler in log.handlers if isinstance(handler, StdoutHandler)]

    original_methods = {}  # Store original methods to restore them later
    for handler in handlers_to_patch:
        original_methods[handler] = handler.send_message
        handler.send_message = new_send_message.__get__(handler)  # Bind method to instance

    try:
        yield captured_messages
    finally:
        # Cleanup: Restore the original send_message methods
        for handler in handlers_to_patch:
            handler.send_message = original_methods[handler]


@pytest.fixture(scope="function")
def cloudwatch():
    json_log._logger = None

    with patch.dict(os.environ, {
        'AWS_DEFAULT_REGION': 'us-west-2',
        'AWS_ACCESS_KEY_ID': 'testing',
        'AWS_SECRET_ACCESS_KEY': 'testing'
    }):
        yield

    if json_log._logger is not None:
        handlers_to_remove = [
            handler for handler in json_log._logger.handlers
            if isinstance(handler, BaseHandler)
        ]
        for handler in handlers_to_remove:
            json_log._logger.removeHandler(handler)
        json_log._logger = None


class MockLogRecord(logging.LogRecord):
    def __init__(self, name, level, filename, lineno, message, args, exc_info, func=None, sinfo=None):
        super().__init__(name, level, filename, lineno, message, args, exc_info, func, sinfo)
        self.message = message
        self.args = args
        self.exc_info = exc_info


class MockOpenSearchClient:
    bulk_calls: List[Tuple[List[Any], Dict[str, Any]]] = []

    def __init__(self):
        self.captured_requests = []
        self.mock_bulk_exception = False
        self.bulk_calls = []

    def ping(self):
        return True

    def index(self, *args, **kwargs):
        self.captured_requests.append((args, kwargs))
        return {}

    @property
    def transport(self):
        transport = MagicMock()
        transport.serializer = OpenSearchSerializer()
        return transport

    def bulk(self, *args, **kwargs):
        self.bulk_calls.append({"actions": []})  # type: ignore
        actions_tuple = args or kwargs.get('body', [])
        response_items = []

        actions = actions_tuple[0].split('\n')
        for i in range(0, len(actions), 2):
            if actions[i].strip():
                action = json.loads(actions[i].strip())
                data = json.loads(actions[i + 1].strip())
                self.bulk_calls[-1]["actions"].append({"_source": data})
                if 'index' in action:
                    data['_id'] = data.get('_id', None)
                    response_items.append({"index": {"status": 200}})
                    self.bulk_calls[-1]["actions"][-1]["_index"] = action['index']['_index']
        if self.mock_bulk_exception:
            self.mock_bulk_exception = False
            raise AssertionError("Mocked exception")
        return {
            "took": 1,
            "errors": False,
            "items": response_items
        }


@pytest.fixture(scope="function")
def opensearch_handler():
    json_log._logger = None

    with patch(
            'opensearch_log.opensearch_handler.OpensearchHandler._get_opensearch_client', return_value=MockOpenSearchClient()):
        handler = OpensearchHandler()
        yield handler
        handler.close()

    if json_log._logger is not None:
        handlers_to_remove = [
            handler for handler in json_log._logger.handlers
            if isinstance(handler, BaseHandler)
        ]
        for handler in handlers_to_remove:
            json_log._logger.removeHandler(handler)
        json_log._logger = None


@pytest.fixture(scope="function")
def cloudwatch_handler():
    json_log._logger = None
    log_group = "test-log-group"
    log_stream = "test-log-stream"

    with patch.dict(os.environ, {
        'AWS_DEFAULT_REGION': 'us-west-2',
        'AWS_ACCESS_KEY_ID': 'testing',
        'AWS_SECRET_ACCESS_KEY': 'testing'
    }), mock_logs():
        # Create the log group and stream beforehand
        client = boto3.client("logs")
        client.create_log_group(logGroupName=log_group)
        client.create_log_stream(logGroupName=log_group, logStreamName=log_stream)

        handler = CloudwatchHandler(log_group=log_group, log_stream=log_stream)
        yield handler
        handler.close()

    if json_log._logger is not None:
        handlers_to_remove = [
            handler for handler in json_log._logger.handlers
            if isinstance(handler, BaseHandler)
        ]
        for handler in handlers_to_remove:
            json_log._logger.removeHandler(handler)
        json_log._logger = None


import datetime
import json

from opensearch_log import Logging, log_fields
from opensearch_log.stdout_handler import get_logger
from tests.conftest import capture_logs


def test_get_logger_with_same_component(caplog, log):
    log1 = get_logger()
    log2 = get_logger()

    assert log1 is log2
    assert log1 is log

    log.info("This is an info message inside context manager.")

    assert len(caplog.records) == 1

    record = caplog.records[0]
    assert record.msg == "This is an info message inside context manager."
    assert getattr(record, "application") == "-mock-application-"


def test_get_logger_with_different_component(caplog):
    log = get_logger("AnotherComponent")
    log.info("This is an info message inside context manager.")

    assert len(caplog.records) == 1

    record = caplog.records[0]
    assert record.msg == "This is an info message inside context manager."
    assert getattr(record, "application") == "AnotherComponent"


def test_log_var_decorator_with_value(caplog, log):
    @log_fields(explicit_field="ExplicitValue")
    def log_message():
        log.info("This is an info message from decorated function with explicit value.")

    log_message()

    assert len(caplog.records) == 1

    record = caplog.records[0]
    assert record.msg == "This is an info message from decorated function with explicit value."
    assert getattr(record, "explicit_field") == "ExplicitValue"


def test_log_ctx_with_values(caplog, log):
    with Logging(explicit_field_ctx="ExplicitValueCtx", field2="Value2"):
        log.info("This is an info message inside context manager with explicit value.")

    assert len(caplog.records) == 1

    record = caplog.records[0]
    assert record.msg == "This is an info message inside context manager with explicit value."
    assert getattr(record, "explicit_field_ctx") == "ExplicitValueCtx"
    assert getattr(record, "field2") == "Value2"


def test_nested_log_var_decorator(caplog, log):
    @log_fields(image_id="Value1InDecorator")
    @log_fields(var2_field="Value2InDecorator")
    @log_fields(var3_field=datetime.datetime.now())
    def log_message():
        log.info("This is an info message from decorated function with a list of LogVars.")

    log_message()

    assert len(caplog.records) == 1

    record = caplog.records[0]
    assert record.msg == "This is an info message from decorated function with a list of LogVars."
    assert getattr(record, "image_id") == "Value1InDecorator"
    assert getattr(record, "var2_field") == "Value2InDecorator"


def test_logger_output_json(log):
    with capture_logs(log) as logs:
        log.info("Mock message")

        log_contents = logs.getvalue().strip()
        log_data = json.loads(log_contents.split("\n")[0])  # pytest add handlers that duplicate log messages

        assert log_data["message"] == "Mock message"
        assert log_data["application"] == "-mock-application-"


def test_default_logger_is_changed(log):
    with capture_logs(log) as logs:
        log.info("Mock message2")

        log_contents = logs.getvalue().strip()
        log_data = json.loads(log_contents.split("\n")[0])  # pytest add handlers that duplicate log messages

        assert log_data["message"] == "Mock message2"
        assert log_data["application"] == "-mock-application-"


def test_get_logger_reinit(log):
    with capture_logs(log) as logs:
        _logger = None

        log = get_logger("AnotherComponent")
        log.info("Reinited.")

        log_contents = logs.getvalue().strip()
        log_data = json.loads(log_contents.split("\n")[0])  # pytest add handlers that duplicate log messages

        assert log_data["message"] == "Reinited."
        assert log_data["application"] == "AnotherComponent"


def test_get_logger_fields(caplog):
    log1 = get_logger(application="test", branch="master", key1="value1", key2="value2")
    log1.info("Hello")
    record = caplog.records[0]
    assert record.msg == "Hello"
    assert getattr(record, "application") == "test"
    assert getattr(record, "branch") == "master"
    assert getattr(record, "key1") == "value1"
    assert getattr(record, "key2") == "value2"

    log2 = get_logger(application="test2", branch="master2", key1="new_value1", key3="value3")
    log2.info("Hello2")
    record = caplog.records[1]
    assert record.msg == "Hello2"
    assert getattr(record, "application") == "test2"
    assert getattr(record, "branch") == "master2"
    assert getattr(record, "key1") == "new_value1"
    assert getattr(record, "key3") == "value3"
    assert not hasattr(record, "key2")



"""JSON logging."""
import logging
import sys
from functools import wraps
from typing import Any, Dict, List, Optional, Type

from pythonjsonlogger import jsonlogger

from opensearch_log.base_handler import BaseHandler

_logger: Optional[logging.Logger] = None
_logger_params: Optional[Dict[str, Optional[str]]] = None
_log_values: Dict[str, Any] = {}


def set_record_factory() -> None:
    """Set a log record factory."""
    original_factory = logging.getLogRecordFactory()

    def _record_factory(*args: Any, **kwargs: Any) -> logging.LogRecord:
        record = original_factory(*args, **kwargs)
        for field, value in _log_values.items():
            setattr(record, field, value)
        return record

    logging.setLogRecordFactory(_record_factory)


def create_logger(
    log_handler_instance: BaseHandler, level: int, clear_handlers: bool = False
) -> logging.Logger:
    """Create a logger that stream logs in JSON format with additional fields."""
    result = logging.getLogger()
    if clear_handlers:
        remove_handlers(result)
    formatter = get_json_formatter()
    log_handler_instance.setFormatter(formatter)
    result.addHandler(log_handler_instance)
    result.setLevel(level)
    return result


def get_json_formatter() -> jsonlogger.JsonFormatter:
    """Get a JSON formatter."""
    return jsonlogger.JsonFormatter(  # type: ignore
        "%(message)%(levelname)%(name)%(application)%(filename)",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )


def replace_logger_fields(fields_to_log: Dict[str, Any]) -> None:
    """Update logger with new fields to log."""
    global _logger_params  # pylint: disable=global-statement
    if _logger_params == fields_to_log or (fields_to_log["application"] is None is None):
        return
    assert (
        _logger_params is not None
    ), "The method should be called only if the logger have been created before"
    remove_log_fields(*_logger_params.keys())
    add_log_fields(**fields_to_log)
    _logger_params = fields_to_log


def remove_handlers(_logger: logging.Logger) -> None:
    """Remove any default handlers."""
    if _logger.handlers:
        for handler in _logger.handlers:
            _logger.removeHandler(handler)


def add_log_fields(**values: Any) -> List[str]:
    """Include fields to all log records."""
    _log_values.update(values)
    return list(values.keys())


def remove_log_fields(*fields: str) -> None:
    """Remove logged fields."""
    for key in fields:
        _log_values.pop(key, None)


class Logging:
    """Context manager to add fields to log records."""

    def __init__(self, **values: Any) -> None:
        """Initialize the context manager with fields to add to log records."""
        self.values = values
        self.added_fields: List[str] = []

    def __enter__(self) -> "Logging":
        """Enter the context: add fields to log records."""
        self.added_fields = add_log_fields(**self.values)
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[Any],
    ) -> None:
        """Exit the context: remove the added fields from log records."""
        remove_log_fields(*self.added_fields)


def log_fields(func: Optional[Any] = None, **values: Any) -> Any:
    """Decorate to include fields `values` to all log records."""

    def decorate(func: Any) -> Any:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            added_fields = add_log_fields(**values)
            try:
                return func(*args, **kwargs)
            finally:
                remove_log_fields(*added_fields)

        return wrapper

    return decorate(func) if callable(func) else decorate


def get_logger(
    application: Optional[str] = sys.argv[0],
    *,
    log_handler: BaseHandler,
    level: int = logging.INFO,
    clear_handlers: bool = False,
    **values: Any,
) -> logging.Logger:
    """Get a JSON logger."""
    global _logger, _logger_params  # pylint: disable=global-statement

    if _logger is not None:
        replace_logger_fields({"application": application, **values})
        return _logger

    _logger_params = {"application": application, **values}
    _logger = create_logger(
        log_handler_instance=log_handler, level=level, clear_handlers=clear_handlers
    )
    add_log_fields(**_logger_params)
    set_record_factory()
    return _logger



"""A Python logging handler for efficient and reliable direct log transmission to OpenSearch."""
from opensearch_log.json_log import Logging, add_log_fields, log_fields, remove_log_fields

__all__ = ["Logging", "add_log_fields", "log_fields", "remove_log_fields"]



"""JSON logging to stdout."""
import logging
from typing import Any, Optional

from opensearch_log import json_log
from opensearch_log.base_handler import BaseHandler
from opensearch_log.json_log import get_json_formatter


class StdoutHandler(BaseHandler):
    """Handler to send json log to stdout."""

    def __init__(self) -> None:
        """Initialize the handler."""
        super().__init__()

    def send_message(self, message: Optional[str], record: logging.LogRecord) -> None:
        """Send the log message to stdout."""
        print(message)


def get_logger(
    application: Optional[str] = None,
    branch: Optional[str] = None,
    *,
    level: int = logging.INFO,
    clear_handlers: bool = False,
    **values: Any,
) -> logging.Logger:
    """Get a logger to send JSON logs to stdout."""
    return json_log.get_logger(
        application=application,
        branch=branch,
        level=level,
        log_handler=StdoutHandler(),
        clear_handlers=clear_handlers,
        **values,
    )


def add_stdout_json_handler(logger: logging.Logger) -> None:
    """Add a stdout handler to the logger."""
    stdout_handler = StdoutHandler()
    stdout_handler.setFormatter(get_json_formatter())
    logger.addHandler(stdout_handler)



import json
import logging

import boto3

from opensearch_log import Logging, json_log
from tests.conftest import MockLogRecord, capture_logs
from opensearch_log.stdout_handler import add_stdout_json_handler


def test_cloudwatch_logging_existing_group_and_stream(cloudwatch_handler):
    logger = json_log.get_logger(
        component="-fake-component-",
        log_handler=cloudwatch_handler,
    )
    test_message = "Test Message for Existing Group and Stream"
    logger.info(test_message)
    cloudwatch_handler.flush()

    # Check if CloudWatch put_log_events was called
    client = boto3.client('logs')
    response = client.get_log_events(
        logGroupName='test-log-group',
        logStreamName='test-log-stream'
    )
    messages = [event['message'] for event in response['events']]
    assert test_message == json.loads(messages[0])["message"]


def test_cloudwatch_emit(cloudwatch_handler):
    message = 'Test log message'
    record = MockLogRecord(
        name='test_logger',
        level=logging.INFO,
        filename='test.py',
        lineno=10,
        message=message,
        args=None,
        exc_info=None
    )

    cloudwatch_handler.emit(record)

    assert len(cloudwatch_handler._buffer) == 1
    assert cloudwatch_handler._buffer[0]['message'] == message


def test_cloudwatch_flush(cloudwatch_handler):
    record = MockLogRecord(
        name='test_logger',
        level=logging.INFO,
        filename='test.py',
        lineno=10,
        message='Test log message',
        args=None,
        exc_info=None
    )

    cloudwatch_handler.emit(record)
    assert len(cloudwatch_handler._buffer) == 1

    cloudwatch_handler.flush()

    assert len(cloudwatch_handler._buffer) == 0


def test_cloudwatch_close(cloudwatch_handler):
    record = MockLogRecord(
        name='test_logger',
        level=logging.INFO,
        filename='test.py',
        lineno=10,
        message='Test log message',
        args=None,
        exc_info=None
    )

    cloudwatch_handler.emit(record)
    assert len(cloudwatch_handler._buffer) == 1
    cloudwatch_handler.close()

    assert len(cloudwatch_handler._buffer) == 0


def test_cloudwatch_handler(cloudwatch_handler):
    logger = json_log.get_logger(
        component="-fake-component-",
        log_handler=cloudwatch_handler,
    )
    with Logging(image_id="-fake-image-"):
        message = "Test log message"
        logger.info(message)
        cloudwatch_handler.flush()

        # Check if CloudWatch put_log_events was called
        client = boto3.client('logs')
        response = client.describe_log_streams(logGroupName='test-log-group')

        # Check if the log stream exists and has stored events
        assert len(response['logStreams']) > 0
        assert 'firstEventTimestamp' in response['logStreams'][0]

        # Check if the specific message was logged
        response = client.get_log_events(
            logGroupName='test-log-group',
            logStreamName='test-log-stream'
        )
        messages = [event['message'] for event in response['events']]
        assert message == json.loads(messages[0])["message"]


def test_cloudwatch_logger_echo_stdout(cloudwatch_handler):
    logger = json_log.get_logger(
        component="-fake-component-",
        log_handler=cloudwatch_handler,
        clear_handlers=True,
    )
    add_stdout_json_handler(logger)

    with capture_logs(logger) as logs:
        logger.info("Test message")

        log_contents = logs.getvalue().strip()
        log_data = json.loads(log_contents.split("\n")[0])

        assert log_data["message"] == "Test message"
        assert log_data["component"] == "-fake-component-"


def test_cloudwatch_logger_do_not_echo_stdout(cloudwatch_handler):
    logger = json_log.get_logger(
        component="-fake-component-",
        log_handler=cloudwatch_handler,
        clear_handlers=True,
    )

    with capture_logs(logger) as logs:
        logger.info("Test message")

        log_contents = logs.getvalue().strip()
        assert log_contents == ''



import json
import logging
from datetime import datetime, timezone
from time import sleep

import pytest

from opensearch_log.opensearch_serializer import OpenSearchSerializer
from unittest.mock import MagicMock, patch

from opensearch_log.opensearch_handler import OpensearchHandler, INDEX_DATE_FORMAT, DEFAULT_INDEX_NAME
from opensearch_log.stdout_handler import add_stdout_json_handler
from opensearch_log import json_log
from opensearch_log import Logging
from tests.conftest import MockLogRecord, opensearch_handler, capture_logs
import opensearchpy.exceptions


def test_emit(opensearch_handler):
    record = MockLogRecord(
        name='test_logger',
        level=logging.INFO,
        filename='test.py',
        lineno=10,
        message='Test log message',
        args=None,
        exc_info=None
    )

    opensearch_handler.emit(record)

    assert len(opensearch_handler._buffer) == 1
    assert opensearch_handler._buffer[0]['message'] == 'Test log message'
    assert opensearch_handler._buffer[0]['filename'] == 'test.py'


def test_flush(opensearch_handler):
    record = MockLogRecord(
        name='test_logger',
        level=logging.INFO,
        filename='test.py',
        lineno=10,
        message='Test log message',
        args=None,
        exc_info=None
    )

    opensearch_handler.emit(record)

    opensearch_handler.flush()

    # Assert that the buffer is empty after flushing
    assert len(opensearch_handler._buffer) == 0


def test_close(opensearch_handler):
    record = MockLogRecord(
        name='test_logger',
        level=logging.INFO,
        filename='test.py',
        lineno=10,
        message='Test log message',
        args=None,
        exc_info=None
    )

    opensearch_handler.emit(record)

    opensearch_handler.close()

    # Assert that the buffer is empty after closing
    assert len(opensearch_handler._buffer) == 0


def test_index_naming(opensearch_handler):
    opensearch_handler.index_rotate = OpensearchHandler.MONTHLY
    index_name = opensearch_handler._get_index_name()

    # Check the prefix
    assert index_name.startswith(DEFAULT_INDEX_NAME)

    # Check the date part of the index name
    current_date = datetime.now(timezone.utc).replace(day=1)
    expected_date_str = current_date.strftime(INDEX_DATE_FORMAT)
    assert index_name.endswith(expected_date_str)

    opensearch_handler.index_rotate = OpensearchHandler.DAILY
    index_name = opensearch_handler._get_index_name()

    # Check the prefix
    assert index_name.startswith(DEFAULT_INDEX_NAME)

    # Check the date part of the index name
    current_date = datetime.now(timezone.utc)
    expected_date_str = current_date.strftime(INDEX_DATE_FORMAT)
    assert index_name.endswith(expected_date_str)


def test_get_opensearch_datetime_str(opensearch_handler):
    timestamp = 1634616000.123  # October 19, 2021, 12:00:00.123 UTC

    # Calculate the expected UTC datetime based on the timestamp
    expected_datetime = datetime.utcfromtimestamp(timestamp)

    # Get the datetime_str from the method
    datetime_str = opensearch_handler._get_opensearch_datetime_str(timestamp)

    # Parse the datetime_str to a datetime object
    actual_datetime = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%fZ')

    assert actual_datetime == expected_datetime


def test_test_opensearch_connection(opensearch_handler):
    mock_opensearch_client = MagicMock()

    with patch.object(mock_opensearch_client, 'ping', return_value=True):
        assert opensearch_handler.is_connected() is True


def test_opensearch_handler_ping_called():
    mock_opensearch_client = MagicMock()

    with patch(
            'opensearch_log.opensearch_handler.OpensearchHandler._get_opensearch_client',
            return_value=mock_opensearch_client):
        OpensearchHandler()
        mock_opensearch_client.ping.assert_called_once()


def test_opensearch_log(opensearch_handler):
    logger = json_log.get_logger(
        application="-mock-component-",
        log_handler=opensearch_handler,
    )
    with Logging(my_field="-mock-my-field-"):
        logger.info("Mock log message")
        opensearch_handler.flush()

    mock_client = opensearch_handler._get_opensearch_client()

    assert len(mock_client.bulk_calls) == 1
    kwargs = mock_client.bulk_calls[0]


    assert kwargs["actions"][0]["_index"] == opensearch_handler._get_index_name()
    assert kwargs["actions"][0]["_source"]["message"] == "Mock log message"
    assert kwargs["actions"][0]["_source"]["my_field"] == "-mock-my-field-"
    assert kwargs["actions"][0]["_source"]["application"] == "-mock-component-"
    assert kwargs["actions"][0]["_source"]["levelname"] == "INFO"


def test_opensearch_log_echo_stdout(opensearch_handler):
    logger = json_log.get_logger(
        application="-mock-component-",
        log_handler=opensearch_handler,
        clear_handlers=True,
    )
    add_stdout_json_handler(logger)

    with capture_logs(logger) as logs:
        logger.info("Mock message")

        log_contents = logs.getvalue().strip()
        log_data = json.loads(log_contents.split("\n")[0])  # pytest add handlers that duplicate log messages

        assert log_data["message"] == "Mock message"
        assert log_data["application"] == "-mock-component-"


def test_opensearch_log_not_echo_stdout(opensearch_handler):
    logger = json_log.get_logger(
        application="-mock-component-",
        log_handler=opensearch_handler,
        echo_stdout=False,
        clear_handlers=True,
    )

    with capture_logs(logger) as logs:
        logger.info("Mock message")

        log_contents = logs.getvalue().strip()
        assert log_contents == ''


def test_opensearch_log_context_unserializable(opensearch_handler):
    """If default OpenSearch serialization fails we catch the exception and use default serialization."""
    class UnSerializable:
        pass

    logger = json_log.get_logger(
        application="-mock-component-",
        log_handler=opensearch_handler,
        echo_stdout=False,
        clear_handlers=True,
    )

    with Logging(field=UnSerializable()):
        logger.info("Mock message")
        logger.info("Mock message2")


def test_opensearch_log_unserializable(opensearch_handler):

    class UnSerializable:
        pass

    assert "UnSerializable" in OpenSearchSerializer().dumps({"field": UnSerializable()})


def test_opensearch_log_flush_exception(opensearch_handler):
    record = MockLogRecord(
        name='test_logger',
        level=logging.INFO,
        filename='test.py',
        lineno=10,
        message='Test log message',
        args=None,
        exc_info=None
    )

    opensearch_handler.flush_seconds = 0.001
    opensearch_handler.flush()
    opensearch_handler._get_opensearch_client().mock_bulk_exception = True
    opensearch_handler._get_opensearch_client().bulk_calls = []
    opensearch_handler.emit(record)
    sleep(0.1)  # wait for timer to call flush()

    assert len(opensearch_handler._get_opensearch_client().bulk_calls) == 2  # exc and retry
    assert len(opensearch_handler._buffer) == 0  # we managed to send in retry

    opensearch_handler.flush_seconds = 100
    opensearch_handler.emit(record)
    assert len(opensearch_handler._buffer) == 1
    opensearch_handler.close()
    assert len(opensearch_handler._buffer) == 0


def test_opensearch_log_bulk_size(opensearch_handler):
    with patch.object(opensearch_handler, 'buffer_size', 1):
        logger = json_log.get_logger(
            application="-mock-component-",
            log_handler=opensearch_handler,
            echo_stdout=False,
            clear_handlers=True,
        )
        logger.info("Mock message")
        logger.info("Mock message2")
        opensearch_handler.flush()
        assert len(opensearch_handler._get_opensearch_client().bulk_calls) == 2

    opensearch_handler._get_opensearch_client().bulk_calls = []

    with patch.object(opensearch_handler, 'buffer_size', 2):
        logger = json_log.get_logger(
            application="-mock-component-",
            log_handler=opensearch_handler,
            echo_stdout=False,
            clear_handlers=True,
        )
        logger.info("Mock message")
        logger.info("Mock message2")
        opensearch_handler.flush()
        assert len(opensearch_handler._get_opensearch_client().bulk_calls) == 1



"""Base logger for structured logging."""
import logging
from typing import Optional


# Common Parent
class BaseHandler(logging.Handler):
    """Abstract base handler for structured logging."""

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a logging record in a processed format."""
        message = self.format(record)
        self.send_message(message, record)

    def send_message(self, message: Optional[str], record: logging.LogRecord) -> None:
        """Send the log message. This method should be implemented by subclasses."""
        raise NotImplementedError("This method should be implemented in subclasses.")
